### System ###
import socket
import pickle
import asyncio
import threading
import traceback
from abc import abstractmethod, ABC

### Logging ###
from logzero import logger

### Data Handling ###
import websockets
from compress_pickle import dumps

### zeroconf ###
from zeroconf import Zeroconf, ServiceInfo, IPVersion


def get_ip_address():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    return s.getsockname()[0]


class PullWorker(threading.Thread, ABC):

    def __init__(self):
        super(PullWorker, self).__init__()
        self.data = {}
        self.last_data = {}
        self.stop_flag = False
        self.name = type(self).__name__

    def run(self):
        logger.info(f"Starting {self.name}")

        try:
            self.execute()
        except Exception as e:
            logger.error(f"Encountered exception in {self.name}: {e}")
            traceback.print_exc()

        logger.info(f"Stopped {self.name}")

    @abstractmethod
    def execute(self):
        pass

    def get(self, ignore_cache=False):
        if ignore_cache and self.data is not None:
            return self.data
        elif self.data is not None and self.last_data is not self.data:
            self.last_data = self.data
            return self.data
        return None


class PushWorker(threading.Thread, ABC):

    SERVICE_NAME = None
    SERVICE_PORT = None
    SERVICE_APPLICATION = None
    SERVICE_CAPABILITIES = []

    def __init__(self, loop):
        super(PushWorker, self).__init__()
        self.loop = loop
        self.data = {}
        self.last_data = {}
        self.stop_flag = False
        self.name = type(self).__name__
        self.service_info = None
        if self.SERVICE_NAME and self.SERVICE_PORT and self.SERVICE_APPLICATION and self.SERVICE_CAPABILITIES:
            self.zeroconf = Zeroconf(ip_version=IPVersion.V4Only)
            self.service_info = ServiceInfo(
                "_http._tcp.local.",
                f"{self.SERVICE_NAME}._http._tcp.local.",
                addresses=[socket.inet_aton(get_ip_address())],
                port=self.SERVICE_PORT,
                properties=pickle.dumps({
                    "application": self.SERVICE_APPLICATION,
                    "capabilities": self.SERVICE_CAPABILITIES
                }, protocol=4))

    def run(self):
        logger.info(f"Starting {self.name}")

        if self.service_info:
            logger.info(f"Registering service {self.SERVICE_NAME}")
            self.zeroconf.register_service(self.service_info)

        try:
            self.execute()
        except Exception as e:
            logger.error(f"Encountered exception in {self.name}: {e}")
            traceback.print_exc()
        finally:
            if self.service_info:
                logger.info(f"Unregistering service {self.SERVICE_NAME}")
                self.zeroconf.unregister_service(self.service_info)
                self.zeroconf.close()

        logger.info(f"Stopped {self.name}")

    @abstractmethod
    def execute(self):
        pass

    async def handler(self, websocket, path):
        self.connected.add(websocket)
        try:
            await websocket.recv()
        except websockets.exceptions.ConnectionClosed:
            pass
        finally:
            self.connected.remove(websocket)

    def send_data(self, data):
        dts = dumps(data, compression="lz4", protocol=4)
        for websocket in self.connected.copy():
            coro = websocket.send(dts)
            _ = asyncio.run_coroutine_threadsafe(coro, self.loop)



### System ###
import time
import atexit
import socket
import pickle
from collections import namedtuple, defaultdict

### zeroconf ###
from zeroconf import IPVersion, ServiceBrowser, Zeroconf, ServiceListener, ServiceInfo


SERVICES = defaultdict(list)
CAPABILITIES = defaultdict(list)
OWN_SERVICES = defaultdict(list)

Info = namedtuple("info", ("properties", "addresses", "service_info"))


def get_ip_address():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    return s.getsockname()[0]


def better_listener_args(func):
    def listener_method_proxy(self, zc, type_, name):
        service_info = zc.get_service_info(type_, name)
        if service_info:
            properties = pickle.loads(service_info.text)
            addresses = [(socket.inet_ntoa(addr), service_info.port) for addr in service_info.addresses]
            info = Info(properties, addresses, service_info)
        else:
            info = Info(None, None, None)
        name = name.rstrip(type_)
        func(self, zc, info, type_, name)
    return listener_method_proxy


class CustomListener(ServiceListener):

    @better_listener_args
    def add_service(self, zc, info, type_, name):
        if info.properties.get("application", None) == "TSM":
            SERVICES[name].append(info)
            capabilities = info.properties.get("capabilities", [])
            for cap in capabilities:
                CAPABILITIES[cap].append(info)

    @better_listener_args
    def remove_service(self, zc, info, type_, name):
        if name in SERVICES:
            SERVICES[name].remove(info)
            capabilities = info.properties.get("capabilities", [])
            for cap in capabilities:
                CAPABILITIES[cap].remove(info)

    @better_listener_args
    def update_service(self, zc, info, type_, name):
        if info.properties.get("application", None) == "TSM":
            SERVICES[name].remove(info)
            SERVICES[name].append(info)
            capabilities = info.properties.get("capabilities", [])
            for cap in capabilities:
                CAPABILITIES[cap].remove(info)
                CAPABILITIES[cap].append(info)


def get_service_by_name(name, block=True):
    while True:
        result = SERVICES.get(name, None)
        if not block or result:
            return list(sorted(result, key=lambda x: x.service_info.priority)) if result else None
        time.sleep(0.5)


def get_service_by_capability(capabilities, block=True):
    # TODO: Implement properly
    # Behavior:
    # Search each service against the desired capabilities
    # We may return more than one service if capabilities are only available split across them
    # For each desired capability, we create a list of services who have that capability
    # Within each list, we sort them by the closeness of the match to the desired capabilities
    # For example, if we desire [gps, imud] and we have two services:
    # S1: [gps]
    # S2: [gps, imud]
    # Then S2 would rank before S1
    # Return Value:
    # {cap: [services] for cap in capabilities}
    # TODO: Should block until all capabilities can be filled

    capabilities = set(capabilities)
    filled_capabilities = set()
    selected_services = {}

    while True:
        if filled_capabilities == capabilities:
            return selected_services
        for cap in capabilities - filled_capabilities:
            result = CAPABILITIES.get(cap)
            if result:
                selected_services[cap] = sorted(result, key=lambda x: x[1], reverse=True)[0]
                filled_capabilities.add(cap)
        if not block:
            return selected_services
        time.sleep(0.5)


def advertise_service(name, application, capabilities, port, address=None, type_="_http._tcp.local.", block=True):
    info = ServiceInfo(f"{type_}", f"{name}.{type_}",
                       addresses=[socket.inet_aton(address if address else get_ip_address())], port=port,
                       properties=pickle.dumps({"application": application, "capabilities": capabilities}, protocol=4))
    OWN_SERVICES[name].append(info)
    ZC.register_service(info)

    while block:
        time.sleep(1.0)


def unadvertise_service(name):
    for service in OWN_SERVICES[name]:
        ZC.unregister_service(service)


def cleanup():
    ZC.close()


atexit.register(cleanup)

ZC = Zeroconf(ip_version=IPVersion.V4Only)

BROWSER = ServiceBrowser(ZC, ["_http._tcp.local."], listener=CustomListener())
